@@ -2,10 +2,11 @@ from __future__ import annotations
 
 from typing import TYPE_CHECKING, Generic, TypeVar
 
-from narwhals._expression_parsing import apply_n_ary_operation
+from narwhals._expression_parsing import ExprKind, ExprNode
 
 if TYPE_CHECKING:
     from narwhals.expr import Expr
+    from narwhals.typing import IntoExpr
 
 ExprT = TypeVar("ExprT", bound="Expr")
 
@@ -38,12 +39,10 @@ class ExprStringNamespace(Generic[ExprT]):
             |└───────┴───────────┘|
             └─────────────────────┘
         """
-        return self._expr._with_elementwise(
-            lambda plx: self._expr._to_compliant_expr(plx).str.len_chars()
-        )
+        return self._expr._append_node(ExprNode(ExprKind.ELEMENTWISE, "str.len_chars"))
 
     def replace(
-        self, pattern: str, value: str | ExprT, *, literal: bool = False, n: int = 1
+        self, pattern: str, value: str | IntoExpr, *, literal: bool = False, n: int = 1
     ) -> ExprT:
         r"""Replace first matching regex/literal substring with a new string value.
 
@@ -67,22 +66,20 @@ class ExprStringNamespace(Generic[ExprT]):
             |1  abc abc123   abc123|
             └──────────────────────┘
         """
-        return self._expr._with_elementwise(
-            lambda plx: (
-                apply_n_ary_operation(
-                    plx,
-                    lambda self, value: self.str.replace(
-                        pattern, value, literal=literal, n=n
-                    ),
-                    self._expr,
-                    value,
-                    str_as_lit=True,
-                )
+        return self._expr._append_node(
+            ExprNode(
+                ExprKind.ELEMENTWISE,
+                "str.replace",
+                value,
+                pattern=pattern,
+                literal=literal,
+                n=n,
+                str_as_lit=True,
             )
         )
 
     def replace_all(
-        self, pattern: str, value: str | ExprT, *, literal: bool = False
+        self, pattern: str, value: IntoExpr, *, literal: bool = False
     ) -> ExprT:
         r"""Replace all matching regex/literal substring with a new string value.
 
@@ -105,17 +102,14 @@ class ExprStringNamespace(Generic[ExprT]):
             |1  abc abc123      123|
             └──────────────────────┘
         """
-        return self._expr._with_elementwise(
-            lambda plx: (
-                apply_n_ary_operation(
-                    plx,
-                    lambda self, value: self.str.replace_all(
-                        pattern, value, literal=literal
-                    ),
-                    self._expr,
-                    value,
-                    str_as_lit=True,
-                )
+        return self._expr._append_node(
+            ExprNode(
+                ExprKind.ELEMENTWISE,
+                "str.replace_all",
+                value,
+                pattern=pattern,
+                literal=literal,
+                str_as_lit=True,
             )
         )
 
@@ -138,8 +132,8 @@ class ExprStringNamespace(Generic[ExprT]):
             ... )
             {'fruits': ['apple', '\nmango'], 'stripped': ['apple', 'mango']}
         """
-        return self._expr._with_elementwise(
-            lambda plx: self._expr._to_compliant_expr(plx).str.strip_chars(characters)
+        return self._expr._append_node(
+            ExprNode(ExprKind.ELEMENTWISE, "str.strip_chars", characters=characters)
         )
 
     def starts_with(self, prefix: str) -> ExprT:
@@ -163,8 +157,8 @@ class ExprStringNamespace(Generic[ExprT]):
             |2   None       None|
             └───────────────────┘
         """
-        return self._expr._with_elementwise(
-            lambda plx: self._expr._to_compliant_expr(plx).str.starts_with(prefix)
+        return self._expr._append_node(
+            ExprNode(ExprKind.ELEMENTWISE, "str.starts_with", prefix=prefix)
         )
 
     def ends_with(self, suffix: str) -> ExprT:
@@ -188,8 +182,8 @@ class ExprStringNamespace(Generic[ExprT]):
             |2   None       None|
             └───────────────────┘
         """
-        return self._expr._with_elementwise(
-            lambda plx: self._expr._to_compliant_expr(plx).str.ends_with(suffix)
+        return self._expr._append_node(
+            ExprNode(ExprKind.ELEMENTWISE, "str.ends_with", suffix=suffix)
         )
 
     def contains(self, pattern: str, *, literal: bool = False) -> ExprT:
@@ -218,9 +212,9 @@ class ExprStringNamespace(Generic[ExprT]):
             default_match: [[true,false,true]]
             case_insensitive_match: [[true,false,true]]
         """
-        return self._expr._with_elementwise(
-            lambda plx: self._expr._to_compliant_expr(plx).str.contains(
-                pattern, literal=literal
+        return self._expr._append_node(
+            ExprNode(
+                ExprKind.ELEMENTWISE, "str.contains", pattern=pattern, literal=literal
             )
         )
 
@@ -247,10 +241,8 @@ class ExprStringNamespace(Generic[ExprT]):
             |2  papaya       ya|
             └──────────────────┘
         """
-        return self._expr._with_elementwise(
-            lambda plx: self._expr._to_compliant_expr(plx).str.slice(
-                offset=offset, length=length
-            )
+        return self._expr._append_node(
+            ExprNode(ExprKind.ELEMENTWISE, "str.slice", offset=offset, length=length)
         )
 
     def split(self, by: str) -> ExprT:
@@ -279,9 +271,7 @@ class ExprStringNamespace(Generic[ExprT]):
             |└─────────┴────────────────┘|
             └────────────────────────────┘
         """
-        return self._expr._with_elementwise(
-            lambda plx: self._expr._to_compliant_expr(plx).str.split(by=by)
-        )
+        return self._expr._append_node(ExprNode(ExprKind.ELEMENTWISE, "str.split", by=by))
 
     def head(self, n: int = 5) -> ExprT:
         r"""Take the first n elements of each string.
@@ -305,8 +295,8 @@ class ExprStringNamespace(Generic[ExprT]):
             lyrics: [["taata","taatatata","zukkyun"]]
             lyrics_head: [["taata","taata","zukky"]]
         """
-        return self._expr._with_elementwise(
-            lambda plx: self._expr._to_compliant_expr(plx).str.slice(0, n)
+        return self._expr._append_node(
+            ExprNode(ExprKind.ELEMENTWISE, "str.slice", offset=0, length=n)
         )
 
     def tail(self, n: int = 5) -> ExprT:
@@ -331,10 +321,8 @@ class ExprStringNamespace(Generic[ExprT]):
             lyrics: [["taata","taatatata","zukkyun"]]
             lyrics_tail: [["taata","atata","kkyun"]]
         """
-        return self._expr._with_elementwise(
-            lambda plx: self._expr._to_compliant_expr(plx).str.slice(
-                offset=-n, length=None
-            )
+        return self._expr._append_node(
+            ExprNode(ExprKind.ELEMENTWISE, "str.slice", offset=-n, length=None)
         )
 
     def to_datetime(self, format: str | None = None) -> ExprT:
@@ -375,8 +363,8 @@ class ExprStringNamespace(Generic[ExprT]):
             |└─────────────────────┘|
             └───────────────────────┘
         """
-        return self._expr._with_elementwise(
-            lambda plx: self._expr._to_compliant_expr(plx).str.to_datetime(format=format)
+        return self._expr._append_node(
+            ExprNode(ExprKind.ELEMENTWISE, "str.to_datetime", format=format)
         )
 
     def to_date(self, format: str | None = None) -> ExprT:
@@ -404,8 +392,8 @@ class ExprStringNamespace(Generic[ExprT]):
             |a: [[2020-01-01,2020-01-02]]|
             └────────────────────────────┘
         """
-        return self._expr._with_elementwise(
-            lambda plx: self._expr._to_compliant_expr(plx).str.to_date(format=format)
+        return self._expr._append_node(
+            ExprNode(ExprKind.ELEMENTWISE, "str.to_date", format=format)
         )
 
     def to_uppercase(self) -> ExprT:
@@ -430,9 +418,7 @@ class ExprStringNamespace(Generic[ExprT]):
             |1   None      None|
             └──────────────────┘
         """
-        return self._expr._with_elementwise(
-            lambda plx: self._expr._to_compliant_expr(plx).str.to_uppercase()
-        )
+        return self._expr._append_node(ExprNode(ExprKind.ELEMENTWISE, "str.to_uppercase"))
 
     def to_lowercase(self) -> ExprT:
         r"""Transform string to lowercase variant.
@@ -451,9 +437,7 @@ class ExprStringNamespace(Generic[ExprT]):
             |1   None      None|
             └──────────────────┘
         """
-        return self._expr._with_elementwise(
-            lambda plx: self._expr._to_compliant_expr(plx).str.to_lowercase()
-        )
+        return self._expr._append_node(ExprNode(ExprKind.ELEMENTWISE, "str.to_lowercase"))
 
     def to_titlecase(self) -> ExprT:
         """Modify strings to their titlecase equivalent.
@@ -462,21 +446,8 @@ class ExprStringNamespace(Generic[ExprT]):
             This is a form of case transform where the first letter of each word is
             capitalized, with the rest of the word in lowercase.
 
-        Warning:
-            Different backends might follow different rules to determine what a "word" is:
-
-            - duckdb, polars and spark-like use non-**alphanumeric** characters to
-                define the word boundaries.
-            - pandas-like, pyarrow and dask use non-**alphabetic** characters to define
-                the word boundaries, matching the behavior of
-                [`str.title`](https://docs.python.org/3/library/stdtypes.html#str.title).
-
-            We can observe the difference with the string `"with123numbers"`:
-
-            - non-**alphanumeric** -> `"With123numbers"`
-                - notice lowercase **n** after the digits
-            - non-**alphabetic** -> `"With123Numbers"`
-                - notice uppercase **N** after the digits
+            Word boundaries are defined by non-**alphabetic** characters, matching the
+            behavior of [`str.title`](https://docs.python.org/3/library/stdtypes.html#str.title)
 
         Examples:
             >>> import polars as pl
@@ -507,9 +478,7 @@ class ExprStringNamespace(Generic[ExprT]):
             |└─────────────────────────┴─────────────────────────┘|
             └─────────────────────────────────────────────────────┘
         """
-        return self._expr._with_elementwise(
-            lambda plx: self._expr._to_compliant_expr(plx).str.to_titlecase()
-        )
+        return self._expr._append_node(ExprNode(ExprKind.ELEMENTWISE, "str.to_titlecase"))
 
     def zfill(self, width: int) -> ExprT:
         """Transform string to zero-padded variant.
@@ -535,6 +504,6 @@ class ExprStringNamespace(Generic[ExprT]):
             |3   None      None|
             └──────────────────┘
         """
-        return self._expr._with_elementwise(
-            lambda plx: self._expr._to_compliant_expr(plx).str.zfill(width)
+        return self._expr._append_node(
+            ExprNode(ExprKind.ELEMENTWISE, "str.zfill", width=width)
         )
